//     Copyright 2017, Kay Hayen, mailto:kay.hayen@gmail.com
//
//     Part of "Nuitka", an optimizing Python compiler that is compatible and
//     integrates with CPython, but also works on its own.
//
//     Licensed under the Apache License, Version 2.0 (the "License");
//     you may not use this file except in compliance with the License.
//     You may obtain a copy of the License at
//
//        http://www.apache.org/licenses/LICENSE-2.0
//
//     Unless required by applicable law or agreed to in writing, software
//     distributed under the License is distributed on an "AS IS" BASIS,
//     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//     See the License for the specific language governing permissions and
//     limitations under the License.
//
/**
 * This is responsible for small helper routines that do path
 * manipulations.
 */

#if defined(_WIN32)
#include <Shlwapi.h>
#else
#include <libgen.h>
#endif

/* This abstracts "dirname" for use on Windows and Unix alike. */
char *getDirname( char *path )
{
#if defined(_WIN32)
    PathRemoveFileSpec( path );
	return path;
#else
	return dirname( path );
#endif
}